unit jassvmhooks;

{$mode objfpc}{$H+}
{$ASMMODE INTEL}

interface

uses
	hooks, wc3main, wc3jassvm;

type
	TInstanceEventCallback = procedure (hInstance: TInstanceHandle);
	TOnInstanceResumeExecution = function (hInstance: TInstanceHandle; ScriptResult: PDword): TScriptExitCode;
	TInstanceExecuteFunctionByNameCallback = function (hInstance: TInstanceHandle; Name: PChar; ScriptResult: PDword): TScriptExitCode;
	TInstanceExecuteFunctionByHandleCallback = function (hInstance: TInstanceHandle; hFunction: Dword; ScriptResult: PDword): TScriptExitCode;
	TInstanceGetStringHandleCallback = function (hInstance: TInstanceHandle; Value: PChar): TJassStringHandle;
	TEnableVarChangeEventCallback = procedure (VariableName: PChar);

var
	pOnInstanceCreated: TInstanceEventCallback;
	pOnInstanceDestroying: TInstanceEventCallback;
	pOnInstanceInit: TInstanceEventCallback;
	pOnInstanceResumeExecution: TOnInstanceResumeExecution;
	pOnExecuteFunctionByName: TInstanceExecuteFunctionByNameCallback;
	pOnExecuteFunctionByHandle: TInstanceExecuteFunctionByHandleCallback;
	pOnInstanceGetStringHandle: TInstanceGetStringHandleCallback;
	pOnEnableVarChangeEvent: TEnableVarChangeEventCallback;

implementation

uses
	wc3observer;

type
	TCGameState_ConnectToVarChangeEvent = function (variable_name: PChar; event_id: Dword; handler: PObserver): LongBool; stdcall; // thiscall

var
	orig_RegisterInstance: function ({Instance: PInstance; DesiredSlot: TInstanceHandle}): TInstanceHandle; stdcall; // fastcall
	orig_FreeInstance: procedure ({InstanceId: TInstanceHandle}); stdcall; // fastcall
	orig_RunInstance: function ({Instance: TInstanceHandle; FuncName: PChar;} out Result: Dword; CallInit, Resume: LongBool; OpLimit: Dword): TScriptExitCode; stdcall; // fastcall
	orig_ExecuteFunctionByHandle: function ({hInstance: TInstanceHandle; hFunction: Dword;} Parameters: Pointer; ScriptResult: PDword; OpLimit: Dword; Unused: Dword; ForceNewRun: LongBool): TScriptExitCode; stdcall; // fastcall
	orig_Instance_GetStringHandle: function ({Instance: PInstance;} Value: PChar): TJassStringHandle; stdcall; // thiscall
	orig_CGameState_ConnectToVarChangeEvent: TCGameState_ConnectToVarChangeEvent;

function hook_RegisterInstance({Instance: PInstance; DesiredSlot: TInstanceHandle}): TInstanceHandle; stdcall;
begin
	Result:= orig_RegisterInstance();
	pOnInstanceCreated(Result);
end;

procedure hook_FreeInstance({InstanceId: TInstanceHandle}); stdcall; // fastcall
var
	hInstance: TInstanceHandle;
begin
	asm mov hInstance, ecx end;

	pOnInstanceDestroying(hInstance);

	asm mov ecx, hInstance end;
	orig_FreeInstance();
end;

function hook_RunInstance({Instance: TInstanceHandle; FuncName: PChar;} ScriptResult: PDword; CallInit, Resume: LongBool; OpLimit: Dword): TScriptExitCode; stdcall; // fastcall
var
	hInstance: TInstanceHandle;
	FuncName: PChar;
	Instance: PInstance;
begin
	asm
		mov hInstance, ecx
		mov FuncName, edx
	end;

	Instance:= GetJassThreadLocal()^.Instances[hInstance];

	if Resume and (Instance^.CurrentOpcode <> nil) then begin
		pOnInstanceResumeExecution(hInstance, ScriptResult);
	end
	else begin
		if CallInit then begin
			pOnInstanceInit(hInstance);
		end;

		Result:= pOnExecuteFunctionByName(hInstance, FuncName, ScriptResult);
	end;
end;

function hook_ExecuteFunctionByHandle({hInstance: TInstanceHandle; hFunction: Dword;} Parameters: Pointer; ScriptResult: PDword; OpLimit: Dword; Unused: Dword; ForceNewRun: LongBool): TScriptExitCode; stdcall; // fastcall
var
	hInstance: TInstanceHandle;
	hFunction: Dword;
begin
	asm
		mov hInstance, ecx
		mov hFunction, edx
	end;

	Result:= pOnExecuteFunctionByHandle(hInstance, hFunction, ScriptResult);
end;

function hook_Instance_GetStringHandle({Instance: PInstance;} Value: PChar): TJassStringHandle; stdcall; // thiscall
var
	this: PInstance;
begin
	asm mov this, ecx end;
	Result:= pOnInstanceGetStringHandle(this^.Handle, Value);
end;

function hook_CGameState_ConnectToVarChangeEvent(variable_name: PChar; event_id: Dword; handler: PObserver): LongBool; stdcall; // thiscall
begin
	Result:= orig_CGameState_ConnectToVarChangeEvent(variable_name, event_id, handler);
	pOnEnableVarChangeEvent(variable_name);
end;

procedure Init;
begin
	Hook(Pointer(_RegisterInstance), @hook_RegisterInstance, Pointer(orig_RegisterInstance));
	Hook(Pointer(_FreeInstance), @hook_FreeInstance, Pointer(orig_FreeInstance));
	Hook(Pointer(_RunInstance), @hook_RunInstance, Pointer(orig_RunInstance));
	Hook(Pointer(_ExecuteFunctionByHandle), @hook_ExecuteFunctionByHandle, Pointer(orig_ExecuteFunctionByHandle));
	Hook(Pointer(_Instance_GetStringHandle), @hook_Instance_GetStringHandle, Pointer(orig_Instance_GetStringHandle));
	Hook(Pointer(GameDll + $4312D0), @hook_CGameState_ConnectToVarChangeEvent, Pointer(orig_CGameState_ConnectToVarChangeEvent));
end;

begin
	wc3main.SubscribeOnInitDone(@Init);
end.

