unit logging;

{$mode ObjFPC}{$H+}

interface

uses
	Classes, SysUtils;

type

	{ TLogger }

	TLogger = class
	private
		FBuffer: String;
		FIndentLevel: Integer;
		FFileHandle: THandle;
	private
		procedure EnsureBufferSize(ASize: Integer);
		procedure IncreaseIndent;
		procedure DecreaseIndent;
	public
		procedure DebugLn(Message: PChar; ALength: Integer);
		procedure DebugLn(Message: String);
		procedure DebugLn(Message: String; Arguments: array of const);
		procedure DebugLnEnter(Message: PChar; ALength: Integer);
		procedure DebugLnEnter(Message: String);
		procedure DebugLnEnter(Message: String; Arguments: array of const);
		procedure DebugLnExit;
		constructor Create(AFilePath: String);
	end;

const
	INDENT_CHAR: Char = #9;

implementation

{ TLogger }

procedure TLogger.EnsureBufferSize(ASize: Integer);
begin
	if Length(FBuffer) < ASize then begin
		SetLength(FBuffer, ASize);
	end;
end;

procedure TLogger.IncreaseIndent;
begin
	Inc(FIndentLevel);
	EnsureBufferSize(FIndentLevel * SizeOf(INDENT_CHAR));
	FBuffer[FIndentLevel]:= INDENT_CHAR;
end;

procedure TLogger.DecreaseIndent;
begin
	if FIndentLevel = 0
		then DebugLn('Attempt to decrease indent level when already equal to 0.')
		else Dec(FIndentLevel);
end;

procedure TLogger.DebugLn(Message: PChar; ALength: Integer);
const
	LINE_ENDING: String = #13#10;
var
	RequiredSize: Integer;
begin
	RequiredSize:= Length(INDENT_CHAR) * FIndentLevel
		+ ALength
		+ Length(LINE_ENDING);

	EnsureBufferSize(RequiredSize);

	Move(Message[0], FBuffer[FIndentLevel + 1], ALength);
	Move(LINE_ENDING[1], FBuffer[FIndentLevel + 1 + ALength], Length(LINE_ENDING));

	FileWrite(FFileHandle, FBuffer[1], RequiredSize);
end;

procedure TLogger.DebugLn(Message: String);
begin
	DebugLn(PChar(Message), Length(Message));
end;

procedure TLogger.DebugLn(Message: String; Arguments: array of const);
begin
	DebugLn(Format(Message, Arguments));
end;

procedure TLogger.DebugLnEnter(Message: PChar; ALength: Integer);
begin
	DebugLn(Message, ALength);
	IncreaseIndent;
end;

procedure TLogger.DebugLnEnter(Message: String);
begin
	DebugLnEnter(PChar(Message), Length(Message));
end;

procedure TLogger.DebugLnEnter(Message: String; Arguments: array of const);
begin
	DebugLnEnter(Format(Message, Arguments));
end;

procedure TLogger.DebugLnExit;
begin
	DecreaseIndent;
end;

constructor TLogger.Create(AFilePath: String);
begin
	FIndentLevel:= 0;
	FFileHandle:= FileCreate(AFilePath, fmShareDenyNone);
	if FFileHandle = feInvalidHandle then begin
		raise Exception.CreateFmt('Failed to create file "%s"', [AFilePath]);
	end;
end;

end.


