unit myjassvm;

{$mode objfpc}{$H+}
{$define PROFILER_COMPATIBILITY}
{$define VANILLA_COMPATIBILITY}
{$define NATIVE_CALLBACKS_SUPPORT}
{$define SET_ALL_TYPE_FIELDS}
{$define JASS_VM_VAR_CHANGE_EVENT_ENABLED}

interface

uses
	{$ifopt D+}LazLoggerBase,{$endif}
	fgl,
	DebugUtils,
	data, pool,
	handletables,
	wc3_native_signature_parser,
	wc3observer, wc3string, wc3game, wc3_game_state, wc3jassvm,
	{$ifdef JASS_VM_FAST_FLOAT}fast_float{$else}wc3float{$endif};

type
	{$ifdef JASS_VM_FAST_FLOAT}
	TJassFloat = TFastFloatWar3;
	PJassFloat = PFastFloatWar3;
	{$else}
	TJassFloat = TFloatWar3;
	PJassFloat = PFloatWar3;
	{$endif}

	TVariablesTable = specialize TDataTable<PScriptData>;

	TNativesDataTable = specialize TDataTable<PNativeFunc>;
	TFunctionsDataTable = specialize TDataTable<PFuncAddr>;

	TVarChangeObserversTable = specialize TDataTable<PObserver>;

	TThreads = class;
	TFrameData = class;

	{ TNativeData }

	TNativeData = class
	public
		ArgumentsType: array of TScriptDataType;
	public
		constructor Create(const ANative: TNativeFunc);
	end;

	TNativeExtraDataTable = specialize TDataTable<TNativeData>;

	{ TFramePoolManager }

	TFramePoolManager = object
	public
		class function CreateInstance({%H-}AParameter: Pointer): PFrame; static;
		class procedure ClearInstance(AInstance: PFrame; {%H-}AParameter: Pointer); static;
		class procedure DestroyInstance(AInstance: PFrame; {%H-}AParameter: Pointer); static;
	end;

	{ TFrameDataPoolManager }

	TFrameDataPoolManager = object
	public
		class function CreateInstance({%H-}AParameter: Pointer): TFrameData; static;
		class procedure ClearInstance(AInstance: TFrameData; {%H-}AParameter: Pointer); static;
		class procedure DestroyInstance(AInstance: TFrameData; {%H-}AParameter: Pointer); static;
	end;

	{ TScriptDataPoolManager }

	TScriptDataPoolManager = object
	public
		class function CreateInstance({%H-}AParameter: Pointer): PScriptData; static;
		class procedure ClearInstance(AInstance: PScriptData; {%H-}AParameter: Pointer); static;
		class procedure DestroyInstance(AInstance: PScriptData; {%H-}AParameter: Pointer); static;
	end;

	{ TJassArrayPoolManager }

	TJassArrayPoolManager = object
	public
		class function CreateInstance({%H-}AParameter: Pointer): PJassArray; static;
		class procedure ClearInstance(AInstance: PJassArray; {%H-}AParameter: Pointer); static;
		class procedure DestroyInstance(AInstance: PJassArray; {%H-}AParameter: Pointer); static;
	end;

	{ TProcessData }

	TProcessData = class
	type
		TScriptDataPool = specialize TPool<PScriptData, TScriptDataPoolManager>;
		TJassArrayPool = specialize TPool<PJassArray, TJassArrayPoolManager>;
		TFramePool = specialize TPool<PFrame, TFramePoolManager>;
		TFrameDataPool = specialize TPool<TFrameData, TFrameDataPoolManager>;
	public
		Instance: PInstance;
		Threads: TThreads;
		ScriptDataPool: TScriptDataPool;
		JassArrayPool: TJassArrayPool;
		FramePool: TFramePool;
		FrameDataPool: TFrameDataPool;
		Natives: TNativesDataTable;
		NativesData: TNativeExtraDataTable;
		Functions: TFunctionsDataTable;
		Globals: TVariablesTable;
		VarChangeObservers: TVarChangeObserversTable;
	public
		constructor Create(AInstance: PInstance);
		destructor Destroy; override;
	end;

	TProcesses = specialize TFPGList<TProcessData>;

	{ TFrameData }

	TFrameData = class
	public
		Locals: TVariablesTable;
	public
		constructor Create;
		procedure Clear;
		destructor Destroy; override;
	end;
	TFrames = specialize TFPGList<TFrameData>;

	{ TThreadData }

	TThreadData = class
	private
		FProcess: TProcessData;
	public
		Frames: TFrames;
	private
		procedure SetProcess(AProcess: TProcessData);
	public
		property Process: TProcessData read FProcess write SetProcess;
	public
		procedure ClearFrames;
		constructor Create(AProcess: TProcessData);
		destructor Destroy; override;
	end;

	TThreads = class(specialize TFPGList<TThreadData>);
	TThreadTable = specialize THandleTable<Pointer>;

	{ TEnvironmentData }

	TEnvironmentData = class
	public
		Processes: TProcesses;
		ThreadTable: TThreadTable;
	public
		constructor Create;
		destructor Destroy; override;
	end;

procedure OnInstanceCreated(hInstance: TInstanceHandle);
procedure OnInstanceDestroying(hInstance: TInstanceHandle);
procedure OnInstanceInit(hInstance: TInstanceHandle);
function OnInstanceResumeExecution(hInstance: TInstanceHandle; ScriptResult: PDword): TScriptExitCode;
function OnInstanceExecuteFunctionByName(hInstance: TInstanceHandle; Name: PChar; ScriptResult: PDword): TScriptExitCode;
function OnInstanceExecuteFunctionByHandle(hInstance: TInstanceHandle; hFunction: Dword; ScriptResult: PDword): TScriptExitCode;
function OnInstanceGetStringHandle(hInstance: TInstanceHandle; Value: PChar): TJassStringHandle;
procedure pOnEnableVarChangeEvent(VariableName: PChar);

type
	TTransformJassCodeHandleFunction = function (hFunction: TJassFunctionHandle): Dword; register;

procedure SetTransformJassCodeHandleFunction(Callback: TTransformJassCodeHandleFunction);
function GetTransformJassCodeHandleFunction: TTransformJassCodeHandleFunction;
procedure ResetTransformJassCodeHandleFunction;

var
	JassEnvironment: TEnvironmentData;

implementation

type
	TScriptDataContainerType = (ctVariable, ctRegister, ctStackItem);

var
	pTransformJassCodeHandle: TTransformJassCodeHandleFunction;

// FIXME: game uses _InitializedType when setting variable value

// procedure OnScriptCreated(hScript: TScriptHandle);
// begin

// end;

// procedure OnScriptDestroyed(hScript: TScriptHandle);
// begin

// end;

procedure DebugProcess({%H-}Process: TProcessData);
begin
	// for i:= 0 to Process.Natives.Count do begin

	// end;
end;

procedure FillProcess(Process: TProcessData; Instance: PInstance);
var
	Environment: PJassThreadLocal;
	Script: PJassScript;
	Strings: PHashedStringTable;
	Key: PChar;
	Hash: Dword;
	Native: PNativeFunc;
	Func: PFuncAddr;
	i: Dword;
begin
	Environment:= GetJassThreadLocal();
	Script:= Instance^.Script;
	Strings:= Script^.Strings;

	Process.Globals.Size:= Strings^.FItems.Length;
	Process.Natives.Size:= Strings^.FItems.Length;
	Process.Functions.Size:= Strings^.FItems.Length;
	Process.NativesData.Size:= Strings^.FItems.Length;

	for i:= 1 to Strings^.FItems.Length-1 do begin
		Key:= Strings^.FItems[i]^.HashKeyData.Key;
		Hash:= Strings^.FItems[i]^.HashKeyData.Hash;

		if Environment^.Natives.Find(Key, Hash, Native) then begin
			Process.Natives[i]:= Native;
			Process.NativesData[i]:= TNativeData.Create(Native^);
		end;

		if Instance^.NameToFuncAddr^.Find(Key, Hash, Func) then begin
			Process.Functions[i]:= Func;
		end;
	end;

	Process.VarChangeObservers.Size:= Strings^.FItems.Length;

	DebugProcess(Process);
end;

// Copies all the items from a script string table to a instance string table,
// that must be empty so handles in both tables will be the same.
procedure SetupStringTable(pInstance: PInstance);
var
	i: Integer;
	hString: Dword;
begin
	with pInstance^ do begin
		// null item already present so skip him
		for i:= 1 to Script^.Strings^.Length - 1 do begin
			hString:= StringTable.GetHandle(Script^.Strings^.Items[i]);
			Inc(StringTable.ItemPointers[hString]^.FReferences);
		end;
	end;
end;

procedure DummyVariableChangedCallback(
	{%H-}Dummy: Pointer;
	{%H-}NewValue: Dword;
	{%H-}Name: PChar;
	{%H-}Parameter: Pointer
); register; // fastcall
begin

end;

procedure DummyManageHandleFunction({hAgent: Dword; Action: TManageHandleAction;} {%H-}Parameter: Pointer); stdcall; // fastcall
begin

end;

procedure OnInstanceCreated(hInstance: TInstanceHandle);
var
	Instance: PInstance;
	Process: TProcessData;
	Thread, ParentThread: TThreadData;
begin
	{$ifopt D+}DebugLn('Instance %d created.', [hInstance]){$endif};
	Instance:= GetJassThreadLocal()^.Instances[hInstance];

	if not Instance^.IsChild then begin
		Instance^.VarChangedCallback:= @DummyVariableChangedCallback;
		Instance^.RefHandleCallback:= @DummyManageHandleFunction;
		SetupStringTable(Instance);
		Process:= TProcessData.Create(Instance);
		FillProcess(Process, Instance);
		JassEnvironment.Processes.Add(Process);
	end
	else begin
		ParentThread:= TTHREADDATA( JassEnvironment.ThreadTable[Instance^.ParentHandle] );
		Process:= ParentThread.Process;
	end;

	Thread:= TThreadData.Create(Process);
	JassEnvironment.ThreadTable[hInstance]:= Pointer(Thread);
	{$ifopt D+}DebugLn('Instance %d processed.', [hInstance]);{$endif}
end;

function IsStringExist(Instance: PInstance; StringHandle: TJassStringHandle): Boolean;
begin
	Result:= StringHandle < Instance^.StringTable.ItemsCount;
end;

procedure OnStringDestroyed(Instance: PInstance; StringHandle: TJassStringHandle); inline;
var
	Process: TProcessData;
begin
	Process:= TThreadData(JassEnvironment.ThreadTable[Instance^.Handle]).Process;

	{$ifopt D+}
	if StringHandle < Instance^.Script^.Strings^.Length then begin
		DebugLn('Warning! Destroying string from a script table. %d.', [
			StringHandle
		]);
	end;
	{$endif}
end;

procedure DecrementStringReferences(Instance: PInstance; StringHandle: TJassStringHandle); inline;
begin
	if StringHandle <> 0 then begin
		{$ifopt D+}
		if not IsStringExist(Instance, StringHandle) then begin
			Crash('Attempt to decrement references of a freed string %d.', [
				StringHandle
			]);
		end;
		{$endif}

		Instance^.DecrementStringReferences(StringHandle);
		if Instance^.StringTable.ItemPointers[StringHandle]^.FReferences = 0
			then OnStringDestroyed(Instance, StringHandle);
	end;
end;

procedure IncrementStringReferences(Instance: PInstance; StringHandle: TJassStringHandle); inline;
begin
	{$ifopt D+}
	if StringHandle <> 0 then begin
		if not IsStringExist(Instance, StringHandle) then begin
			Crash('Attempt to increment references of a freed string %d.', [
				StringHandle
			]);
		end;
	end;
	{$endif}

	Instance^.IncrementStringReferences(StringHandle);
end;

function GetStringHandle(Instance: PInstance; Value: PChar): TJassStringHandle;
var
	Process: TProcessData;
begin
	if Value <> nil then begin
		Result:= Instance^.StringTable.GetHandle(Value);
		IncrementStringReferences(Instance, Result);
	end
	else begin
		Result:= 0;
	end;
end;

function GetStringObject(Instance: PInstance; StringHandle: TJassStringHandle): PJassString; inline;
begin
	{$ifopt D+}
	if not IsStringExist(Instance, StringHandle) then begin
		Crash('Attempt to get content of a freed string %d.', [
			StringHandle
		]);
	end;
	{$endif}

	Result:= Instance^.StringTable.ItemPointers[StringHandle];
end;

function GetStringContent(Instance: PInstance; StringHandle: TJassStringHandle): PChar; inline;
begin
	Result:= GetStringObject(Instance, StringHandle)^.Content;
end;

procedure DecrementValueReferences(Instance: PInstance; Value: Dword; ValueType: TScriptDataType; TargetContainerType: TScriptDataContainerType); inline;
var
	IsVariable: Boolean;
begin
	IsVariable:= not (TargetContainerType in [ctRegister, ctStackItem]);

	case ValueType of
		stString: DecrementStringReferences(Instance, Value);
		stHandle: {$ifdef VANILLA_COMPATIBILITY}if IsVariable
			then{$endif} Instance^.DecrementHandleReferences(Value);
	end;
end;

procedure IncrementValueReferences(Instance: PInstance; Value: Dword; ValueType: TScriptDataType; TargetContainerType: TScriptDataContainerType); inline;
var
	IsVariable: Boolean;
begin
	IsVariable:= not (TargetContainerType in [ctRegister, ctStackItem]);

	case ValueType of
		stString: IncrementStringReferences(Instance, Value);
		stHandle: {$ifdef VANILLA_COMPATIBILITY}if IsVariable
			then{$endif} Instance^.IncrementHandleReferences(Value);
	end;
end;

procedure ChangeValue(Instance: PInstance; var Value: Dword; NewValue: Dword; ValueType: TScriptDataType; TargetContainerType: TScriptDataContainerType); inline;
begin
	IncrementValueReferences(Instance, NewValue, ValueType, TargetContainerType);
	DecrementValueReferences(Instance, Value, ValueType, TargetContainerType);
	Value:= NewValue;
end;

procedure DecrementScriptDataValueReferences(Instance: PInstance; var Data: TScriptData; TargetContainerType: TScriptDataContainerType); inline;
begin
	DecrementValueReferences(Instance, Data.Value, Data._Type, TargetContainerType);
end;

procedure IncrementScriptDataValueReferences(Instance: PInstance; var Data: TScriptData; TargetContainerType: TScriptDataContainerType); inline;
begin
	IncrementValueReferences(Instance, Data.Value, Data._Type, TargetContainerType);
end;

procedure FreeInstanceArray(Instance: PInstance; var Data: TScriptData); inline;
var
	JassArray: PJassArray;
	i: Integer;
	Thread: TThreadData;
begin
	JassArray:= PJassArray(Data.Value);

	case Data.InitializedType of
		stArrayOfString: begin
			for i:= 0 to JassArray^.List.Length - 1 do begin
				DecrementStringReferences(Instance, JassArray^.List[i]);
			end;
		end;

		{$ifndef VANILLA_COMPATIBILITY}
		stArrayOfHandle: begin
			for i:= 0 to JassArray^.List.Length - 1 do begin
				Instance^.DecrementHandleReferences(JassArray^.List[i]);
			end;
		end;
		{$endif}
	end;

	Thread:= TThreadData(JassEnvironment.ThreadTable[Instance^.Handle]);
	Thread.Process.JassArrayPool.FreeItem(JassArray);
end;

procedure ClearScriptData(Instance: PInstance; var Data: TScriptData; DestinationContainerType: TScriptDataContainerType); inline;
begin
	if (DestinationContainerType = ctVariable)
		and (Data.InitializedType >= stArrayOfInteger)
	then begin
		FreeInstanceArray(Instance, Data);
	end
	else begin
		DecrementScriptDataValueReferences(Instance, Data, DestinationContainerType);
	end;

	Data.Value:= 0;
	Data._Type:= stVoid;

	{$ifndef SET_ALL_TYPE_FIELDS}if DestinationContainerType = ctVariable
		then{$endif} Data.InitializedType:= stVoid;
end;

procedure SetScriptDataValue(Instance: PInstance; var AData: TScriptData; ANewValue: Dword; ANewType: TScriptDataType; DestinationContainerType: TScriptDataContainerType); inline;
begin
	// Increment before decrement in case if Source = Destination
	// to prevent object destruction
	IncrementValueReferences(Instance, ANewValue, ANewType, DestinationContainerType);
	DecrementScriptDataValueReferences(Instance, AData, DestinationContainerType);

	AData.Value:= ANewValue;

	{$ifndef SET_ALL_TYPE_FIELDS}if DestinationContainerType <> ctVariable
		then{$endif} AData._Type:= ANewType;

	{$ifndef SET_ALL_TYPE_FIELDS}if DestinationContainerType = ctVariable
		then{$endif} AData.InitializedType:= ANewType;
end;

procedure InitScriptDataValue(Instance: PInstance; var AData: TScriptData; ANewValue: Dword; ANewType: TScriptDataType; DestinationContainerType: TScriptDataContainerType); inline;
begin
	IncrementValueReferences(Instance, ANewValue, ANewType, DestinationContainerType);

	AData.Value:= ANewValue;

	{$ifndef SET_ALL_TYPE_FIELDS}if DestinationContainerType <> ctVariable
		then{$endif} AData._Type:= ANewType;

	{$ifndef SET_ALL_TYPE_FIELDS}if DestinationContainerType = ctVariable
		then{$endif} AData.InitializedType:= ANewType;
end;

procedure CopyScriptData(Instance: PInstance; const Source: TScriptData; var Destination: TScriptData; DestinationContainerType: TScriptDataContainerType); inline;
begin
	SetScriptDataValue(
		Instance,
		Destination,
		Source.Value,
		Source._Type,
		DestinationContainerType
	);
end;

procedure MoveScriptData(Instance: PInstance; var Source: TScriptData; var Destination: TScriptData; DestinationContainerType: TScriptDataContainerType); inline;
begin
	DecrementScriptDataValueReferences(
		Instance,
		Destination,
		DestinationContainerType
	);

	Destination.Value:= Source.Value;
	Destination._Type:= Source._Type;

	{$ifndef SET_ALL_TYPE_FIELDS}
	if DestinationContainerType = ctVariable then begin
	{$endif}
		Destination.InitializedType:= Source._Type;
	{$ifndef SET_ALL_TYPE_FIELDS}
	end;
	{$endif}

	if DestinationContainerType <> ctStackItem then begin
		Source.Value:= 0;
		Source._Type:= stVoid;
	end;
end;

procedure CleanupInstanceFrames(pInstance: PInstance);
var
	Frame: PFrame;
	Thread: TThreadData;
begin
	Thread:= TThreadData(JassEnvironment.ThreadTable[pInstance^.Handle]);

	while True do begin;
		Frame:= pInstance^.Frames.First;
		if Frame = nil then break;

		pInstance^.Frames.Remove(Frame);
		Thread.Process.FramePool.FreeItem(Frame);
	end;

	Thread.ClearFrames;
end;

procedure ClearInstanceRegisters(Instance: PInstance);
var
	i: Integer;
begin
	for i:= 0 to High(Instance^.Registers) do begin
		ClearScriptData(Instance, Instance^.Registers[i], ctRegister);
	end;
end;

// Detect the memory hack hacked array
function IsJassArrayValid(JassArray: PJassArray): Boolean;
begin
	Result:= JassArray^.List.Capacity <= JASS_MAX_ARRAY_SIZE;
end;

procedure FreeArrayMemory(Variable: PScriptData);
var
	JassArray: PJassArray;
begin
	if (Variable^._Type >= stArrayOfInteger)
		and (Variable^.Value <> 0)
	then begin
		JassArray:= PJassArray(Variable^.Value);
		if IsJassArrayValid(JassArray) then begin
			JassArray^.Done;
			Dispose(JassArray);
		end;
	end;
end;

procedure FreeProcessGlobals(Process: TProcessData);
var
	Variable: PScriptData;
begin
	for Variable in Process.Globals do begin
		FreeArrayMemory(Variable);
		Variable^.Done;
		Dispose(Variable);
	end;
	Process.Instance^.Globals^.Clear;
	Process.Globals.Clear;
end;

procedure UpdateProcessMainThread(Process: TProcessData);
var
	env: PJassThreadLocal;
	OldMainInstance, Instance: PInstance;
	i: Integer;
begin
	OldMainInstance:= Process.Instance;
	Process.Instance:= nil;
	env:= GetJassThreadLocal();

	// When the main thread is destroyed, its status is passed to the first
	// child thread it encounters.
	for i:= 0 to env^.Instances.Length - 1 do begin
		Instance:= env^.Instances[i];
		if Instance = nil then continue;

		if Instance^.ParentHandle = OldMainInstance^.Handle then begin
			Process.Instance:= Instance;
			break;
		end;
	end;
end;

procedure OnInstanceDestroying(hInstance: TInstanceHandle);
var
	Instance: PInstance;
	Process: TProcessData;
	Thread: TThreadData;
begin
	{$ifopt D+}DebugLn('Destroying instance %d.', [hInstance]);{$endif}
	Instance:= GetJassThreadLocal()^.Instances[hInstance];

	Thread:= TTHREADDATA( JassEnvironment.ThreadTable[hInstance] );
	Process:= Thread.Process;

	CleanupInstanceFrames(Instance);
	ClearInstanceRegisters(Instance);
	JassEnvironment.ThreadTable.FreeHandle(hInstance);
	Thread.Destroy;
	
	if Process.Threads.Count = 0 then begin
		FreeProcessGlobals(Process);
		JassEnvironment.Processes.Remove(Process);
		Process.Destroy;
	end
	else if not Instance^.IsChild then begin
		UpdateProcessMainThread(Process);
	end;
	{$ifopt D+}DebugLn('Instance %d data cleared.', [hInstance]);{$endif}
end;

type
	PPHashedString = ^PHashedString;

	TOpDataPack = record
		Instance: PInstance;
		Instruction: PInstruction;

		Frame: PFrame;
		Locals: TInstance.PVariables;
		Globals: TInstance.PVariables;
		ScriptStrings: PPHashedString;

		Code: PInstruction;
		Registers: TInstance.PRegisters;

		Process: TProcessData;
		Thread: TThreadData;
		FastGlobals: TVariablesTable;
		FastLocals: TVariablesTable;
		FastFunctions: TFunctionsDataTable;
		FastNatives: TNativesDataTable;
		FastNativesExtra: TNativeExtraDataTable;
		VarChangeObservers: TVarChangeObserversTable;

		ExitCode: TScriptExitCode;
	end;

	TVariableScope = (vsGlobal, vsLocal);

function GetVariable(hName: TJassStringHandle; const Params: TOpDataPack; out VarScope: TVariableScope): PScriptData; inline;
begin
	with Params do begin
		if FastLocals.GetValue(hName, Result) then begin
			VarScope:= vsLocal;
		end
		else begin
			VarScope:= vsGlobal;
			Result:= FastGlobals.GetValue(hName);
		end;
	end;
end;

function GetVariable(hName: TJassStringHandle; const Params: TOpDataPack): PScriptData; inline;
var
	Unused: TVariableScope;
begin
	Result:= GetVariable(hName, Params, Unused);
end;

procedure PutVariableToMap
(
	hName: TJassStringHandle;
	Variable: PScriptData;
	Map: TInstance.PVariables;
	Strings: PPHashedString
);

inline;

begin
	with Strings[hName]^.HashKeyData do begin
		Map^.FastPut(Key, Hash, Variable);
	end;
end;

function DefineVariable
(
	const Params: TOpDataPack;
	hName: TJassStringHandle;
	VarType: TScriptDataType;
	VarScope: TVariableScope
): PScriptData;

inline;

var
	FastTable: TVariablesTable;
	{$ifdef PROFILER_COMPATIBILITY}
	SlowTable: TInstance.PVariables;
	{$endif}

begin
	Result:= Params.Process.ScriptDataPool.GetItem;
	Result^._Type:= VarType;

	if Integer(VarType) >= Integer(stArrayOfInteger) then begin
		Result^.InitializedType:= VarType;
	end;

	case VarScope of
		vsGlobal:
		begin
			FastTable:= Params.FastGlobals;
			{$ifdef PROFILER_COMPATIBILITY}
			SlowTable:= Params.Globals;
			{$endif}
		end;

		vsLocal:
		begin
			FastTable:= Params.FastLocals;
			{$ifdef PROFILER_COMPATIBILITY}
			SlowTable:= Params.Locals;
			{$endif}
		end;
	end;

	if VarType >= stArrayOfInteger then begin
		PJassArray(Result^.Value):= Params.Process.JassArrayPool.GetItem;
	end;

	FastTable[hName]:= Result;
	{$ifdef PROFILER_COMPATIBILITY}
	PutVariableToMap(hName, Result, SlowTable, Params.ScriptStrings);
	{$endif}
end;

procedure StartFrame(var Params: TOpDataPack); inline;
begin
	with Params do begin
		Frame:= Params.Process.FramePool.GetItem;
		Instance^.Frames.Prepend(Frame);
		Locals:= @Frame^.Locals;

		if Thread.Frames.Last <> nil then begin
			Thread.Frames.Last.Locals:= FastLocals;
		end;
		Thread.Frames.Add(Process.FrameDataPool.GetItem);
		FastLocals:= Thread.Frames.Last.Locals;
	end;
end;

procedure OnFrameChanged(var Params: TOpDataPack); inline;
begin
	with Params do begin
		Frame:= Instance^.Frames.First;

		if Frame <> nil then begin
			Locals:= @Frame^.Locals;
		end;

		if Thread.Frames.Last <> nil then begin
			FastLocals:= Thread.Frames.Last.Locals;
		end;
	end;
end;

procedure EndFrame(var Params: TOpDataPack); inline;
begin
	with Params do begin
		Instance^.Frames.Remove(Frame);
		Process.FramePool.FreeItem(Frame);
		Process.FrameDataPool.FreeItem(Thread.Frames.Last);
		Thread.Frames.Count:= Thread.Frames.Count - 1;
		OnFrameChanged(Params);
	end;
end;

procedure OP_DefineLocal(const Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('local %s', [
				Instance^.Script^.Strings^[Parameter]
			]);
			{$endif}
			DefineVariable(Params, Parameter, TScriptDataType(r1), vsLocal);
		end;
	end;
end;

procedure OP_DefineGlobal(const Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('global %s', [
				Instance^.Script^.Strings^[Parameter]
			]);
			{$endif}
			DefineVariable(Params, Parameter, TScriptDataType(r1), vsGlobal);
		end;
	end;
end;

procedure OP_DefineArgument(const Params: TOpDataPack); inline;
var
	Variable: PScriptData;
begin
	with Params do begin
		with Instruction^ do begin
			Variable:= DefineVariable(Params, Parameter, TScriptDataType(r1), vsLocal);
			Variable^.IsArgument:= True;

			with Instance^.Frames.GetNextItem(Frame)^ do begin
				{$ifopt D+}
				DebugLn('argument %s = %d', [
					Instance^.Script^.Strings^[Parameter],
					PInt32(@Stack.Items[Stack.Length - 1 - r2]^.Value)^
				]);
				{$endif}

				SetScriptDataValue(
					Instance,
					Variable^,
					Stack.Items[Stack.Length - 1 - r2]^.Value,
					TScriptDataType(r1),
					ctVariable
				);
			end;
		end;
	end;
end;

procedure PushValueToFrame(Instance: PInstance; Frame: PFrame; AValue: Dword; AType: TScriptDataType); inline;
begin
	with Frame^ do begin
		InitScriptDataValue(
			Instance,
			Stack.Items[Stack.Length]^,
			AValue,
			AType,
			ctStackItem
		);
		Inc(Stack.Length);
	end;
end;

procedure PopFrameValue(Instance: PInstance; Frame: PFrame; var Destination: TScriptData; DestinationContainerType: TScriptDataContainerType); inline;
begin
	with Frame^ do begin
		Dec(Stack.Length);
		MoveScriptData(
			Instance,
			Stack.Items[Stack.Length]^,
			Destination,
			DestinationContainerType
		);
	end;
end;

procedure PopFrameValues(Instance: PInstance; Frame: PFrame; Count: Integer); inline;
var
	i: Integer;
begin
	with Frame^ do begin
		for i:= 0 to Count - 1 do begin
			Dec(Stack.Length);
			DecrementScriptDataValueReferences(
				Instance,
				Stack.Items[Stack.Length]^,
				ctStackItem
			);
		end;
	end;
end;

procedure OP_CleanStack(const Params: TOpDataPack); inline;
begin
	with Params do begin
		{$ifopt D+}
		DebugLn('cleanstack %d', [
			Instruction^.r1
		]);
		{$endif}
		PopFrameValues(Instance, Frame, Instruction^.r1);
	end;
end;

procedure OP_Literal(const Params: TOpDataPack); inline;
begin
	with Params do begin
		{$ifopt D+}
		DebugLn('literal r%d = %d', [
			Instruction^.r1,
			PInt32(@Instruction^.Parameter)^
		]);
		{$endif}

		SetScriptDataValue(
			Instance,
			Registers^[Instruction^.r1],
			Instruction^.Parameter,
			TScriptDataType(Instruction^.r2),
			ctRegister
		);
	end;
end;

procedure OP_Move(const Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('move r%d = r%d', [
				r2,
				r1
			]);
			{$endif}
			CopyScriptData(Instance, Registers^[r2], Registers^[r1], ctRegister);
		end;
	end;
end;

function OP_GetVar(var Params: TOpDataPack): Boolean; inline;
var
	Variable: PScriptData;
begin
	Result:= True;

	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('getvar %s -> r%d', [
				Instance^.Script^.Strings^[Parameter],
				r1
			]);
			{$endif}
			// TODO: Add optional null pointer check
			Variable:= GetVariable(Parameter, Params);

			if Variable^.InitializedType <> stVoid then begin
				SetScriptDataValue(
					Instance,
					Registers^[r1],
					Variable^.Value,
					TScriptDataType(r2),
					ctRegister
				);
			end
			else begin
				{$ifopt D+}
				DebugLn('Access to uninitialized variable "%s".', [
					Instance^.Script^.Strings^[Parameter]
				]);
				{$endif}
				ExitCode:= scOverflowOrUndefined;
				Result:= False;
			end;
		end;
	end;
end;

procedure OP_GetCode(const Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('getcode %s -> r%d', [
				Instance^.Script^.Strings^[Parameter],
				r1
			]);
			{$endif}
			// TODO: Add optional null pointer check
			SetScriptDataValue(
				Instance,
				Registers^[r1],
				PtrUInt(FastFunctions[Parameter]^.Address),
				TScriptDataType(r2),
				ctRegister
			);
		end;
	end;
end;

procedure OP_GetArray(const Params: TOpDataPack); inline;
var
	SourceVariable, DestinationRegister, IndexRegister: PScriptData;
	Items: PJassArray;
	Value: Dword = 0;
begin
	with Params do begin
		{$ifopt D+}
		DebugLn('getarray %s -> r%d', [
			Instance^.Script^.Strings^[Instruction^.Parameter],
			Instruction^.r1
		]);
		{$endif}

		DestinationRegister:= @Registers^[Instruction^.r1];
		SourceVariable:= GetVariable(Instruction^.Parameter, Params);
		IndexRegister:= @Registers^[Instruction^.r2];
		Items:= PJassArray(SourceVariable^.Value);

        // TODO: Dont forget to initialize array to prevent null pointer!
		// TODO: Add option to enable runtime range checks

		if IndexRegister^.Value < Items^.List.Length then begin
			Value:= Items^.List.FItems[IndexRegister^.Value]
		end;

		SetScriptDataValue(
			Instance,
			DestinationRegister^,
			Value,
			TScriptDataType(Instruction^.r3),
			ctRegister
		);
	end;
end;

procedure OP_SetVar(const Params: TOpDataPack); inline;
var
	Variable: PScriptData;
	{$ifdef JASS_VM_VAR_CHANGE_EVENT_ENABLED}
	Changed: Boolean = False;
	Observer: PObserver;
	Event: TGameVarChangedEvent;
	VarScope: TVariableScope;
	{$endif}
begin
	with Params do begin
		with Instruction^ do begin
			// TODO: Add optional null pointer check
			// TODO: should I handle null type?
			{$ifopt D+}
			DebugLn('setvar %s = %d', [
				Instance^.Script^.Strings^[Parameter],
				PInt32(@Registers^[r1].Value)^
			]);
			{$endif}

			{$ifdef JASS_VM_VAR_CHANGE_EVENT_ENABLED}
			if (Instance^.VarChangedCallback <> nil)
				and (VarChangeObservers.GetValue(Parameter, Observer))
			then begin
				Variable:= GetVariable(Parameter, Params, VarScope);
				Changed:= Registers^[r1].Value <> Variable^.Value;
				CopyScriptData(Instance, Registers^[r1], Variable^, ctVariable);
				if (VarScope = vsGlobal) and Changed then begin
					{$ifdef JASS_VM_VAR_CHANGE_EVENT_SLOW}
					Instance^.VarChangedCallback(
						nil,
						Variable^.Value,
						Variable^.HashKeyData.Key,
						Instance^.VarChangedCallbackParam
					);
					{$else}
					Event.Id:= $80268;
					Event.NewValue:= PSingle(@Variable^.Value)^;
					Observer^.Emit(Event);
					{$endif}
				end;
			end
			else begin
			{$endif}
				Variable:= GetVariable(Parameter, Params);
				CopyScriptData(Instance, Registers^[r1], Variable^, ctVariable);
			{$ifdef JASS_VM_VAR_CHANGE_EVENT_ENABLED}
			end;
			{$endif}
		end;
	end;
end;

function ExtractArrayItemType(AArrayType: TScriptDataType): TScriptDataType;
begin
	Result:= TScriptDataType(Dword(AArrayType) - 5);
end;

procedure OP_SetArray(const Params: TOpDataPack); inline;
var
	SourceRegister, DestinationVariable, IndexRegister: PScriptData;
	Items: PJassArray;
begin
	with Params do begin
		IndexRegister:= @Registers^[Instruction^.r1];
		SourceRegister:= @Registers^[Instruction^.r2];
		DestinationVariable:= GetVariable(Instruction^.Parameter, Params);
		Items:= PJassArray(DestinationVariable^.Value);

		if IndexRegister^.Value < Items^.List.Length then begin
			// TODO: Dont forget to initialize array to prevent null pointer!
			// TODO: Add option to enable runtime range checks
			{$ifopt D+}
			DebugLn('setarray %s[%d] = %d', [
				Instance^.Script^.Strings^[Instruction^.Parameter],
				PInt32(@IndexRegister^.Value)^,
				PInt32(@SourceRegister^.Value)^
			]);
			{$endif}

			ChangeValue(
				Instance,
				Items^.List.FItems[IndexRegister^.Value],
				SourceRegister^.Value,
				ExtractArrayItemType(DestinationVariable^._Type),
				ctVariable
			);
			// TODO: use callback when global variable changed
		end;
	end;
end;

procedure OP_Push(const Params: TOpDataPack); inline;
begin
	with Params do begin
		// TODO: Dont forget to allocate all items of stack
		// TODO: add optional range check
		{$ifopt D+}
		DebugLn('push r%d', [
			Instruction^.r1
		]);
		{$endif}

		with Registers^[Instruction^.r1] do begin
			PushValueToFrame(Instance, Frame, Value, _Type);
		end;
	end;
end;

procedure OP_Pop(const Params: TOpDataPack); inline;
begin
	with Params do begin
		// TODO: Dont forget to allocate all items of stack
		// TODO: add optional range check
		{$ifopt D+}
		DebugLn('pop r%d', [
			Instruction^.r1
		]);
		{$endif}

		PopFrameValue(Instance, Frame, Registers^[Instruction^.r1], ctRegister);
	end;
end;

function ASM_CallNative(pNative: Pointer; Arguments: Pointer; Count: Dword): Dword; assembler; stdcall; inline;
asm
	push esi
	push edi

	mov ecx, Count

	mov eax, ecx
	shl eax, 2
	sub esp, eax

	mov esi, Arguments
	mov edi, esp
	cld
	repe movsd

	mov esi, eax

	call [pNative]

	add esp, esi

	pop edi
	pop esi
end;

function OP_NativeCall(var Params: TOpDataPack): Boolean; inline;
const
	// With a return address pushed equals to a maximum stack capacity.
	MAX_ARGUMENTS = 31;
var
	NativeFunction: PNativeFunc;
	NativeExtraData: TNativeData;
	CurrentArgument: PScriptData;
	ConvertedValue: Dword;
	i: Integer;
	CallResult: Dword;
	Arguments: array [0..MAX_ARGUMENTS] of Dword;
	Strings: array [0..MAX_ARGUMENTS] of TRCString;
	StringsCount: Integer = 0;
begin
	Result:= True;

	with Params do begin
		NativeFunction:= FastNatives[Instruction^.Parameter];
		NativeExtraData:= FastNativesExtra[Instruction^.Parameter];

		with NativeExtraData do begin
			for i:= 0 to High(ArgumentsType) do begin
				CurrentArgument:= Frame^.Stack.Items[
					Frame^.Stack.Length - Length(ArgumentsType) + i
				];

				case ArgumentsType[i] of
					stCode:
					begin
						if CurrentArgument^.Value <> 0 then begin
							ConvertedValue:= Instance^.CodeTable.GetHandle(
								PInstruction(CurrentArgument^.Value)
							);
							{$ifdef NATIVE_CALLBACKS_SUPPORT}
							ConvertedValue:= pTransformJassCodeHandle(ConvertedValue);
							{$endif}
						end
						else begin
							ConvertedValue:= 0;
						end;
					end;

					stReal:
					begin
						ConvertedValue:= Dword(@CurrentArgument^.Value);
					end;

					stString:
					begin
						Strings[StringsCount].Init;
						Strings[StringsCount].Content:= Instance^
							.StringTable
							.ItemPointers[
								CurrentArgument^.Value
							]^.Content;

						ConvertedValue:= Dword(@Strings[StringsCount]);

						Inc(StringsCount);
					end;

					else begin
						ConvertedValue:= CurrentArgument^.Value;
					end;
				end;

				Arguments[i]:= ConvertedValue;
			end;

			{$ifopt D+}
			DebugLn('call native %s', [
				NativeFunction^.HashKeyData.Key
			]);
			{$endif}

			CallResult:= ASM_CallNative(
				NativeFunction^.Address,
				@Arguments[0],
				Length(ArgumentsType)
			);

			for i:= 0 to StringsCount - 1 do begin
				Strings[i].Done;
			end;

			// Dont increment references - native function already did it
			DecrementScriptDataValueReferences(
				Instance,
				Registers^[0],
				ctRegister
			);
			with Registers^[0] do begin
				Value:= CallResult;
				_Type:= NativeFunction^.RetType;
			end;

			PopFrameValues(Instance, Frame, Length(ArgumentsType));
		end;

		if Instance^.Sleeping then begin
			ExitCode:= scSleeping;
			Result:= False;
		end
		else if Instance^.Syncing then begin
			ExitCode:= scSyncing;
			Result:= False;
		end;
	end;
end;

procedure OP_Call(var Params: TOpDataPack); inline;
begin
	with Params do begin
		{$ifopt D+}
		DebugLnEnter('call %s', [FastFunctions[Instruction^.Parameter]^.HashKeyData.Key]);
		{$endif}

		PushValueToFrame(
			Instance,
			Frame,
			Dword(Pointer(Instruction + 1) - Pointer(Code)) div SizeOf(Dword),
			stRetAddr
		);

		StartFrame(Params);

		Instance^.CurrentOpcode:= FastFunctions[Instruction^.Parameter]^.Address;
	end;
end;

function SingleAsDword(Value: Single): Dword; inline;
begin
	PSingle(@Result)^:= Value;
end;

procedure OP_I2R(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('i2r r%d', [
				r1
			]);
			{$endif}

			SetScriptDataValue(
				Instance,
				Registers^[r1],
				SingleAsDword(TJassFloat(PInt32(@Registers^[r1].Value)^).Value),
				stReal,
				ctRegister
			);
		end;
	end;
end;

procedure OP_And(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('and r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			SetScriptDataValue(
				Instance,
				Registers^[r1],
				Dword(PBoolean32(@Registers^[r2].Value)^
					and PBoolean32(@Registers^[r3].Value)^),
				stBoolean,
				ctRegister
			);
		end;
	end;
end;

procedure OP_Or(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('or r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			SetScriptDataValue(
				Instance,
				Registers^[r1],
				Dword(PBoolean32(@Registers^[r2].Value)^
					or PBoolean32(@Registers^[r3].Value)^),
				stBoolean,
				ctRegister
			);
		end;
	end;
end;

procedure OP_Equal(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('equal r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stReal: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(
							Boolean32(PJassFloat(@Registers^[r2].Value)^
								.IsApproximatelyEqualTo(
									PJassFloat(@Registers^[r3].Value)^
								)
							)
						),
						stBoolean,
						ctRegister
					);
				end;

				stBoolean: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(PBoolean32(@Registers^[r2].Value)^
							= PBoolean32(@Registers^[r3].Value)^),
						stBoolean,
						ctRegister
					);
				end;

				else begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(Registers^[r2].Value
							= Registers^[r3].Value)),
						stBoolean,
						ctRegister
					);
				end;
			end;
		end;
	end;
end;

procedure OP_NotEqual(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('not_equal r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stBoolean: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(PBoolean32(@Registers^[r2].Value)^
							<> PBoolean32(@Registers^[r3].Value)^),
						stBoolean,
						ctRegister
					);
				end;

				else begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(Registers^[r2].Value
							<> Registers^[r3].Value)),
						stBoolean,
						ctRegister
					);
				end;
			end;
		end;
	end;
end;

procedure OP_LessEqual(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('less_equal r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stInteger: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(PInt32(@Registers^[r2].Value)^
							<= PInt32(@Registers^[r3].Value)^)),
						stBoolean,
						ctRegister
					);
				end;

				stReal: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(PJassFloat(@Registers^[r2].Value)^
							<= PJassFloat(@Registers^[r3].Value)^)),
						stBoolean,
						ctRegister
					);
				end;
			end;
		end;
	end;
end;

procedure OP_GreaterEqual(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('greater_equal r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stInteger: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(PInt32(@Registers^[r2].Value)^
							>= PInt32(@Registers^[r3].Value)^)),
						stBoolean,
						ctRegister
					);
				end;

				stReal: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(PJassFloat(@Registers^[r2].Value)^
							>= PJassFloat(@Registers^[r3].Value)^)),
						stBoolean,
						ctRegister
					);
				end;
			end;
		end;
	end;
end;

procedure OP_Less(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('less r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stInteger: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(PInt32(@Registers^[r2].Value)^
							< PInt32(@Registers^[r3].Value)^)),
						stBoolean,
						ctRegister
					);
				end;

				stReal: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(PJassFloat(@Registers^[r2].Value)^
							< PJassFloat(@Registers^[r3].Value)^)),
						stBoolean,
						ctRegister
					);
				end;
			end;
		end;
	end;
end;

procedure OP_Greater(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('greater r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stInteger: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(PInt32(@Registers^[r2].Value)^
							> PInt32(@Registers^[r3].Value)^)),
						stBoolean,
						ctRegister
					);
				end;

				stReal: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(Boolean32(PJassFloat(@Registers^[r2].Value)^
							> PJassFloat(@Registers^[r3].Value)^)),
						stBoolean,
						ctRegister
					);
				end;
			end;
		end;
	end;
end;

function ConcatenateStrings(a, b: PChar): String;
const
	NULL_STRING = '';
	EMPTY_STRING = ''#0;
var
	LengthA, LengthB, TotalLength: SizeInt;
begin
	if (a = nil) and (b = nil) then begin
		Result:= NULL_STRING;
	end
	else begin
		LengthA:= StrLen(a);
		LengthB:= StrLen(b);
		TotalLength:= LengthA + LengthB;

		if TotalLength > 0 then begin
			SetLength(Result, TotalLength);
			if LengthA > 0 then Move(a^, Result[1], LengthA);
			if LengthB > 0 then Move(b^, Result[1 + LengthA], LengthB);
		end
		else begin
			Result:= EMPTY_STRING;
		end;
	end;
end;

procedure OP_Add(var Params: TOpDataPack); inline;
var
	ResultString: String;
	StringHandle: Dword;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('add r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stInteger: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(PInt32(@Registers^[r2].Value)^
							+ PInt32(@Registers^[r3].Value)^),
						stInteger,
						ctRegister
					);
				end;

				stReal: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						SingleAsDword(
							Single(
								PJassFloat(@Registers^[r2].Value)^
								+ PJassFloat(@Registers^[r3].Value)^
							)
						),
						stReal,
						ctRegister
					);
				end;

				stString: begin
					ResultString:= ConcatenateStrings(
						GetStringContent(Instance, Registers^[r2].Value),
						GetStringContent(Instance, Registers^[r3].Value)
					);

					StringHandle:= GetStringHandle(
						Instance,
						PChar(ResultString)
					);

					DecrementScriptDataValueReferences(
						Instance,
						Registers^[r1],
						ctRegister
					);
					// GetStringHandle function already incremented reference counter
					with Registers^[r1] do begin
						Value:= StringHandle;
						_Type:= stString;
					end;
				end;
			end;
		end;
	end;
end;

procedure OP_Sub(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('sub r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stInteger: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(PInt32(@Registers^[r2].Value)^
							- PInt32(@Registers^[r3].Value)^),
						stInteger,
						ctRegister
					);
				end;

				stReal: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						SingleAsDword(
							Single(
								PJassFloat(@Registers^[r2].Value)^
								- PJassFloat(@Registers^[r3].Value)^
							)
						),
						stReal,
						ctRegister
					);
				end;
			end;
		end;
	end;
end;

procedure OP_Mul(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('mul r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stInteger: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(PInt32(@Registers^[r2].Value)^
							* PInt32(@Registers^[r3].Value)^),
						stInteger,
						ctRegister
					);
				end;

				stReal: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						SingleAsDword(
							Single(
								PJassFloat(@Registers^[r2].Value)^
								* PJassFloat(@Registers^[r3].Value)^
							)
						),
						stReal,
						ctRegister
					);
				end;
			end;
		end;
	end;
end;

function OP_Div(var Params: TOpDataPack): Boolean; inline;
begin
	Result:= True;

	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('div r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			if Registers^[r3].Value = 0 then begin
				ExitCode:= scZeroDivision;
				Exit(False);
			end;

			case Registers^[r2]._Type of
				stInteger: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						Dword(PInt32(@Registers^[r2].Value)^
							div PInt32(@Registers^[r3].Value)^),
						stInteger,
						ctRegister
					);
				end;

				stReal: begin
					SetScriptDataValue(
						Instance,
						Registers^[r1],
						SingleAsDword(
							Single(
								PJassFloat(@Registers^[r2].Value)^
								/ PJassFloat(@Registers^[r3].Value)^
							)
						),
						stReal,
						ctRegister
					);
				end;
			end;
		end;
	end;
end;

procedure OP_Modulo(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('mod r%d, r%d -> r%d', [
				r2,
				r3,
				r1
			]);
			{$endif}

			case Registers^[r2]._Type of
				stInteger: begin
					if Registers^[r3].Value <> 0 then begin
						SetScriptDataValue(
							Instance,
							Registers^[r1],
							Dword(PInt32(@Registers^[r2].Value)^
								mod PInt32(@Registers^[r3].Value)^),
							stInteger,
							ctRegister
						);
					end
					else begin
						SetScriptDataValue(
							Instance,
							Registers^[r1],
							0,
							stInteger,
							ctRegister
						);
					end;
				end;
			end;
		end;
	end;
end;

procedure OP_Negate(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('negate r%d', [
				r1
			]);
			{$endif}

			with Registers^[r1] do begin
				case _Type of
					stInteger: begin
						PInt32(@Value)^:= -PInt32(@Value)^;
					end;

					stReal: begin
						PJassFloat(@Value)^:= -PJassFloat(@Value)^;
					end;
				end;
			end;
		end;
	end;
end;

procedure OP_Not(var Params: TOpDataPack); inline;
begin
	with Params do begin
		with Instruction^ do begin
			{$ifopt D+}
			DebugLn('not r%d', [
				r1
			]);
			{$endif}

			SetScriptDataValue(
				Instance,
				Registers^[r1],
				Dword(not PBoolean32(@Registers^[r1].Value)^),
				stBoolean,
				ctRegister
			);
		end;
	end;
end;

function OP_Return(var Params: TOpDataPack): Boolean; inline;
var
	FinalFrame, ShouldStop: Boolean;
begin
	{$ifopt D+}DebugLn('return');{$endif}

	with Params do begin
		FinalFrame:= Frame^.IsFirstFrame;
		EndFrame(Params);

		ShouldStop:= FinalFrame or (Frame = nil);

		if not ShouldStop then begin
			with Frame^ do begin
				{$ifopt D+}
				with Stack.Items[Stack.Length-1]^ do begin
					if _Type <> stRetAddr then begin
						Crash('Excepted return address but got type(%d).', [_Type]);
					end;
				end;
				{$endif}
				Instance^.CurrentOpcode:= Pointer(Code) + (Stack.Items[Stack.Length-1]^.Value * SizeOf(Dword));
				Dec(Stack.Length);
			end;
		end;

		Result:= not ShouldStop;
	end;

	{$ifopt D+}DebugLnExit;{$endif}
end;

procedure OP_JumpIfTrue(var Params: TOpDataPack); inline;
begin
	with Params do begin
		{$ifopt D+}
		DebugLn('jump_if r%d to %d', [
			Instruction^.r1,
			PInt32(@Instruction^.Parameter)^
		]);
		{$endif}
		if Registers^[Instruction^.r1].Value <> 0 then begin
			// TODO: Add optional range check
			Instance^.CurrentOpcode:= Instance^.Labels^[Instruction^.Parameter];
		end;
	end;
end;

procedure OP_JumpIfFalse(var Params: TOpDataPack); inline;
begin
	with Params do begin
		{$ifopt D+}
		DebugLn('jump_if_not r%d to %d', [
			Instruction^.r1,
			PInt32(@Instruction^.Parameter)^
		]);
		{$endif}
		if Registers^[Instruction^.r1].Value = 0 then begin
			// TODO: Add optional range check
			Instance^.CurrentOpcode:= Instance^.Labels^[Instruction^.Parameter];
		end;
	end;
end;

procedure OP_Jump(var Params: TOpDataPack); inline;
begin
	with Params do begin
		{$ifopt D+}
		DebugLn('jump to %d', [
			PInt32(@Instruction^.Parameter)^
		]);
		{$endif}
		// TODO: Add optional range check
		Instance^.CurrentOpcode:= Instance^.Labels^[Instruction^.Parameter];
	end;
end;

procedure InitializeParameters(out Params: TOpDataPack; AInstance: PInstance);
var
	MyThread: TThreadData;
begin
	MyThread:= TTHREADDATA( JassEnvironment.ThreadTable[AInstance^.Handle] );

	with Params do begin
		Instance:= AInstance;
		// Instruction: initialized at start of each iteration
		Frame:= nil;
		Locals:= nil;
		Globals:= AInstance^.Globals;
		ScriptStrings:= AInstance^.Script^.Strings^.FItems.ItemPointers[0];
		Code:= AInstance^.Script^.Code;
		Registers:= @AInstance^.Registers;
		Process:= MyThread.Process;
		Thread:= MyThread;
		FastGlobals:= Process.Globals;
		// FastLocals: initialized on frame creation
		FastFunctions:= Process.Functions;
		FastNatives:= Process.Natives;
		FastNativesExtra:= Process.NativesData;
		VarChangeObservers:= Process.VarChangeObservers;
	end;
end;

procedure FinalizeParameters(var Params: TOpDataPack);
begin
	with Params do begin
		Process.Globals:= FastGlobals;
		Process.Functions:= FastFunctions;
		Process.Natives:= FastNatives;
		Process.NativesData:= FastNativesExtra;
	end;
end;

procedure BeginExecution(const Params: TOpDataPack);
begin
	with GetJassThreadLocal()^ do begin
		Inc(RunningInstancesCount);

		with RunningInstances do begin
			RunningInstances.Length:= RunningInstancesCount;
			RunningInstances[RunningInstancesCount-1]:= Params.Instance;
		end;
	end;
end;

procedure EndExecution(const Params: TOpDataPack);
begin
	with GetJassThreadLocal()^ do begin
		Dec(RunningInstancesCount);
	end;
end;

procedure StartExecution(var Params: TOpDataPack);
{$ifndef DISABLE_JASS_OPERATION_LIMIT}
var
	ExecutedInstructions: Dword = 0;
	OperationsLimit: Dword;
{$endif}
begin
	BeginExecution(Params);

	with Params.Instance^ do begin
		Sleeping:= False;
		Syncing:= False;
	end;

	{$ifndef DISABLE_JASS_OPERATION_LIMIT}
	OperationsLimit:= Params.Instance^.OperationsLimit;
	{$endif}
	Params.ExitCode:= scExecutionCompleted;

	while True do begin
		{$ifndef DISABLE_JASS_OPERATION_LIMIT}
		if ExecutedInstructions >= OperationsLimit then begin
			Params.ExitCode:= scOpLimitReached;
			break;
		end;
		{$endif}

		Params.Instruction:= Params.Instance^.CurrentOpcode;
		Inc(Params.Instance^.CurrentOpcode);

		case TOperationCode(Params.Instruction^.id) of
			opDebug: OP_Jump(Params);
			opLocal: OP_DefineLocal(Params);
			opGlobal, opConstant: OP_DefineGlobal(Params);
			opArgument: OP_DefineArgument(Params);
			opCleanStack: OP_CleanStack(Params);
			opLiteral: OP_Literal(Params);
			opMove: OP_Move(Params);
			opGetVar: if not OP_GetVar(Params) then break;
			opGetCode: OP_GetCode(Params);
			opGetArray: OP_GetArray(Params);
			opSetVar: OP_SetVar(Params);
			opSetArray: OP_SetArray(Params);
			opPush: OP_Push(Params);
			opPop: OP_Pop(Params);
			opNativeCall: if not OP_NativeCall(Params) then break;
			opCall: OP_Call(Params);
			opI2R: OP_I2R(Params);
			opAnd: OP_And(Params);
			opOr: OP_Or(Params);
			opEqual: OP_Equal(Params);
			opNotEqual: OP_NotEqual(Params);
			opLesserEqual: OP_LessEqual(Params);
			opGreaterEqual: OP_GreaterEqual(Params);
			opLesser: OP_Less(Params);
			opGreater: OP_Greater(Params);
			opAdd: OP_Add(Params);
			opSub: OP_Sub(Params);
			opMul: OP_Mul(Params);
			opDiv: if not OP_Div(Params) then break;
			opModulo: OP_Modulo(Params);
			opNegate: OP_Negate(Params);
			opNot: OP_Not(Params);
			opReturn: if not OP_Return(Params) then break;
			opJumpIfTrue: OP_JumpIfTrue(Params);
			opJumpIfFalse: OP_JumpIfFalse(Params);
			opJump: OP_Jump(Params);
		end;

		{$ifndef DISABLE_JASS_OPERATION_LIMIT}
		Inc(ExecutedInstructions);
		{$endif}
	end;

	EndExecution(Params);
end;

function ResumeExecution(Instance: PInstance; ScriptResult: PDword): TScriptExitCode;
var
	Params: TOpDataPack;
begin
	InitializeParameters(Params, Instance);
	OnFrameChanged(Params);
	StartExecution(Params);
	FinalizeParameters(Params);
	Result:= Params.ExitCode;
	if ScriptResult <> nil then begin
		ScriptResult^:= Instance^.Registers[0].Value;
	end;
end;

function ExecuteCode(pInstance: PInstance; pCode: PInstruction; ScriptResult: PDword = nil): TScriptExitCode;
var
	Params: TOpDataPack;
begin
	pInstance^.CurrentOpcode:= pCode;
	pInstance^.OperationsLimit:= DEFAULT_JASS_OPERATION_LIMIT;

	InitializeParameters(Params, pInstance);
	StartFrame(Params);
	StartExecution(Params);
	Result:= Params.ExitCode;
	FinalizeParameters(Params);

	if ScriptResult <> nil then begin
		ScriptResult^:= pInstance^.Registers[0].Value;
	end;
end;

procedure OnInstanceInit(hInstance: TInstanceHandle);
var
	Instance: PInstance;
	InitFunc: PInitFuncAddr;
	ExitCode: TScriptExitCode;
begin
	{$ifopt D+}
	DebugLnEnter('Initializing instance %d.', [
		hInstance
	]);
	{$endif}
	Instance:= GetJassThreadLocal()^.Instances[hInstance];

	for InitFunc in Instance^.InitFuncs^ do begin
		ExitCode:= ExecuteCode(Instance, InitFunc^.Address);
		CleanupInstanceFrames(Instance);
		{$ifopt D+}DebugLn('ExitCode: %d;', [ExitCode]);{$endif}
	end;
	{$ifopt D+}DebugLnExit();{$endif}
end;

function OnInstanceResumeExecution(hInstance: TInstanceHandle;
	ScriptResult: PDword): TScriptExitCode;
var
	Instance: PInstance;
begin
	Instance:= GetJassThreadLocal()^.Instances[hInstance];
	Result:= ResumeExecution(Instance, ScriptResult);
end;

function OnInstanceExecuteFunctionByName(hInstance: TInstanceHandle;
	Name: PChar; ScriptResult: PDword): TScriptExitCode;
var
	Instance: PInstance;
	JassFunction: PFuncAddr;
begin
	{$ifopt D+}
	DebugLnEnter('Executing function "%s".', [
		Name
	]);
	{$endif}
	Instance:= GetJassThreadLocal()^.Instances[hInstance];

	if Instance^.NameToFuncAddr^.Find(Name, JassFunction) then begin
		Result:= ExecuteCode(Instance, JassFunction^.Address, ScriptResult);
		{$ifopt D+}
		DebugLnExit('ExitCode: %d', [
			Integer(Result)
		]);
		{$endif}
	end
	else begin
		{$ifopt D+}
		DebugLnExit('Function "%s" not found.', [
			Name
		]);
		{$endif}
		Result:= scFunctionNotFound;
	end;
end;

function GetCodeName(Instance: PInstance; Code: PInstruction): String;
var
	FunctionLabel: PInstruction;
begin
	FunctionLabel:= Code - 1;
	if TOperationCode(FunctionLabel^.id) = opFunction then begin
		Result:= Instance^.Script^.Strings^[FunctionLabel^.Parameter];
	end
	else begin
		Result:= 'Instruction at 0x' +  HexStr(Code);
	end;
end;

function OnInstanceExecuteFunctionByHandle(hInstance: TInstanceHandle;
	hFunction: Dword; ScriptResult: PDword): TScriptExitCode;
var
	Instance: PInstance;
	Code: PInstruction;
begin
	Instance:= GetJassThreadLocal()^.Instances[hInstance];

	// If the current instruction is not nil and ForceNewRun flag was not set
	// then resume execution of a previously running code.
	// ForceNewRun is never true so ignore them?
	if Instance^.CurrentOpcode <> nil then begin
		{$ifopt D+}
		DebugLnEnter('Resuming execution of instance %d.', [Instance^.Handle]);
		{$endif}
		Result:= ResumeExecution(Instance, ScriptResult);
		{$ifopt D+}
		DebugLnExit('ExitCode: %d', [
			Integer(Result)
		]);
		{$endif}
	end
	else begin
		Code:= Instance^.CodeTable[hFunction];
		{$ifopt D+}
		DebugLnEnter('Executing function "%s" by handle %d.', [
			GetCodeName(Instance, Code),
			hFunction
		]);
		{$endif}
		Result:= ExecuteCode(Instance, Code, ScriptResult);
		{$ifopt D+}
		DebugLnExit('ExitCode: %d; Result: %d;', [
			Integer(Result),
			ScriptResult^
		]);
		{$endif}
	end;
end;

function OnInstanceGetStringHandle(hInstance: TInstanceHandle; Value: PChar
	): TJassStringHandle;
var
	Instance: PInstance;
begin
	Instance:= GetJassThreadLocal()^.Instances[hInstance];
	Result:= GetStringHandle(Instance, Value);
end;

procedure pOnEnableVarChangeEvent(VariableName: PChar);
var
	Observer: PObserver;
	Process: TProcessData;
	Index: TInstance.TStringTable.TItemHandle;
begin
	Observer:= TGameWar3.Instance^.State^.GetVarChangeEventPublisher(VariableName);
	for Process in JassEnvironment.Processes do begin
		Index:= Process.Instance^.StringTable.FindHandle(VariableName);
		if Index <> 0 then begin
			Process.VarChangeObservers[Index]:= Observer;
		end;
	end;
end;

procedure SetTransformJassCodeHandleFunction(Callback: TTransformJassCodeHandleFunction
	);
begin
	pTransformJassCodeHandle:= Callback;
end;

function GetTransformJassCodeHandleFunction: TTransformJassCodeHandleFunction;
begin
	Result:= pTransformJassCodeHandle;
end;

function DummyTransformJassCodeHandle(hFunction: TJassFunctionHandle): Dword; register;
begin
	Result:= hFunction;
end;

procedure ResetTransformJassCodeHandleFunction;
begin
	pTransformJassCodeHandle:= @DummyTransformJassCodeHandle;
end;

{ TFrameDataPoolManager }

class function TFrameDataPoolManager.CreateInstance(AParameter: Pointer
	): TFrameData;
begin
	Result:= TFrameData.Create;
	Result.Locals.Size:= TProcessData(AParameter).Instance^.Script^.Strings^.Length;
end;

class procedure TFrameDataPoolManager.ClearInstance(AInstance: TFrameData;
	AParameter: Pointer);
var
	Variable: PScriptData;
begin
	with AInstance do begin
		for Variable in Locals do begin
			{$ifdef VANILLA_COMPATIBILITY}
			// Let them leak
			case Variable^._Type of
				stHandle: if not Variable^.IsArgument
					then Variable^._Type:= stInteger;
				stArrayOfHandle: Variable^._Type:= stArrayOfInteger;
			end;
			{$endif}

			TProcessData(AParameter).ScriptDataPool.FreeItem(Variable);
		end;

		Locals.Clear;
	end;
end;

class procedure TFrameDataPoolManager.DestroyInstance(AInstance: TFrameData;
	AParameter: Pointer);
begin
	AInstance.Destroy;
end;

{ TJassArrayPoolManager }

class function TJassArrayPoolManager.CreateInstance(AParameter: Pointer
	): PJassArray;
begin
	New(Result);
	with Result^ do begin
		Init;
		List.Length:= JASS_MAX_ARRAY_SIZE;
		FillByte(List.ItemPointers[0]^, List.Length, 0);
	end;
end;

class procedure TJassArrayPoolManager.ClearInstance(AInstance: PJassArray;
	AParameter: Pointer);
begin
	with AInstance^ do begin
		// There is no way to handle references due to lack of a type information
		FillChar(List.ItemPointers[0]^, List.Length, 0);
	end;
end;

class procedure TJassArrayPoolManager.DestroyInstance(AInstance: PJassArray;
	AParameter: Pointer);
begin
	AInstance^.Done;
	Dispose(AInstance);
end;

{ TScriptDataPoolManager }

class function TScriptDataPoolManager.CreateInstance(AParameter: Pointer
	): PScriptData;
begin
	New(Result);
	Result^.Init;
end;

class procedure TScriptDataPoolManager.ClearInstance(AInstance: PScriptData;
	AParameter: Pointer);
begin
	ClearScriptData(
		PInstance(AParameter),
		AInstance^,
		ctVariable
	);
	AInstance^.IsArgument:= False;
end;

class procedure TScriptDataPoolManager.DestroyInstance(AInstance: PScriptData;
	AParameter: Pointer);
begin
	AInstance^.Done;
	Dispose(AInstance);
end;

{ TFramePoolManager }

class function TFramePoolManager.CreateInstance(AParameter: Pointer): PFrame;
begin
	New(Result);
	Result^.Init;
end;

class procedure TFramePoolManager.ClearInstance(AInstance: PFrame;
	AParameter: Pointer);
begin
	with AInstance^ do begin
		Node.Clear;

		PopFrameValues(
			PInstance(AParameter),
			AInstance,
			AInstance^.Stack.Length
		);

		// A variables are owned by a data table, so dont free them here.
		Locals.Clear;
		IsFirstFrame:= False;
	end;
end;

class procedure TFramePoolManager.DestroyInstance(AInstance: PFrame;
	AParameter: Pointer);
begin
	AInstance^.Done;
	Dispose(AInstance);
end;

{ TNativeData }

constructor TNativeData.Create(const ANative: TNativeFunc);
var
	ParsedSignature: TNativeFunctionSignature;
	i: Integer;
begin
	if ParseNativeFunctionSignature(
		ANative.Signature,
		ParsedSignature,
		False
	)
	then begin
		with ParsedSignature do begin
			SetLength(ArgumentsType, Length(Arguments));

			for i:= 0 to High(Arguments) do begin
				if not GetScriptDataTypeByLetter(
					Arguments[i].TypeId,
					ArgumentsType[i]
				)
				then begin
					ArgumentsType[i]:= stVoid;
				end;
			end;
		end;
	end
	else begin
		Crash('Failed to parse signature of native function "%s%s".', [
			ANative.HashKeyData.Key,
			ANative.Signature
		]);
	end;
end;

{ TProcessData }

constructor TProcessData.Create(AInstance: PInstance);
begin
	Instance:= AInstance;
	Threads:= TThreads.Create;
	ScriptDataPool:= TScriptDataPool.Create(Pointer(Instance));
	JassArrayPool:= TJassArrayPool.Create;
	FramePool:= TFramePool.Create(AInstance);
	FrameDataPool:= TFrameDataPool.Create(Pointer(Self));
	Natives:= TNativesDataTable.Create;
	NativesData:= TNativeExtraDataTable.Create;
	Functions:= TFunctionsDataTable.Create;
	Globals:= TVariablesTable.Create;
	VarChangeObservers:= TVarChangeObserversTable.Create;
end;

destructor TProcessData.Destroy;
var
	Variable: PScriptData;
begin
	Threads.Destroy;
	ScriptDataPool.Destroy;
	JassArrayPool.Destroy;
	FramePool.Destroy;
	FrameDataPool.Destroy;
	Natives.Destroy;
	NativesData.Destroy;
	Functions.Destroy;

	for Variable in Globals do begin
		Variable^.Done;
		Dispose(Variable);
	end;
	Globals.Destroy;

	inherited Destroy;
end;

{ TFrameData }

constructor TFrameData.Create;
begin
	Locals:= TVariablesTable.Create;
end;

procedure TFrameData.Clear;
begin
	Locals.Clear;
end;

destructor TFrameData.Destroy;
begin
	Locals.Destroy;

	inherited Destroy;
end;

{ TThreadData }

procedure TThreadData.SetProcess(AProcess: TProcessData);
begin
	if FProcess <> nil then
		FProcess.Threads.Remove(Self);

	FProcess:= AProcess;

	if FProcess <> nil then
		FProcess.Threads.Add(Self)
end;

procedure TThreadData.ClearFrames;
var
	FrameData: TFrameData;
begin
	for FrameData in Frames do begin
		Process.FrameDataPool.FreeItem(FrameData);
	end;
	Frames.Clear;
end;

constructor TThreadData.Create(AProcess: TProcessData);
begin
	FProcess:= nil;
	Process:= AProcess;
	Frames:= TFrames.Create;
end;

destructor TThreadData.Destroy;
begin
	ClearFrames;
	Frames.Destroy;

	Process:= nil;

	inherited Destroy;
end;

{ TEnvironmentData }

constructor TEnvironmentData.Create;
begin
	Processes:= TProcesses.Create;
	ThreadTable:= TThreadTable.Create;
end;

destructor TEnvironmentData.Destroy;
begin
	Processes.Destroy;
	ThreadTable.Destroy;

	inherited Destroy;
end;

initialization
	ResetTransformJassCodeHandleFunction;
	JassEnvironment:= TEnvironmentData.Create;

finalization
	JassEnvironment.Destroy;

end.

